import socket
import zlib
import time
import thread
import select
import sys

#__HTTP_REQ__ = "GET /~kborders/%s HTTP/1.0\r\n\r\n"
#__SERVER_ADDRESS__ = "141.213.77.24"
__HTTP_REQ__ = "GET /%s HTTP/1.0\r\n\r\n"
__SERVER_ADDRESS__ = "127.0.0.1"
__SERVER_PORT__ = 80
__MAX_RECV__ = 0x400 * 8    # 8 KB
__MAX_IDLE_SECONDS__ = 10    # Number of seconds before timeout

# Test 1:
__RUN_TEST_1__ = False
__SINGLE_FILE_NAME__ = "File500KB.htm"
__SINGLE_FILE_SIZE__ = 0x400 * 500    # 500 KB
__SINGLE_FILE_CHECKSUM__ = -0x7CFDE128
#__SINGLE_FILE_NAME__ = "Test1File500k.htm"
#__SINGLE_FILE_SIZE__ = 100
#__SINGLE_FILE_CHECKSUM__ = 0x6A94C23E

# Test 2:
__RUN_TEST_2__ = False
#__SMALL_FILE_NAME__ = "index.htm"
#__SMALL_FILE_SIZE__ = 0x400
#__SMALL_FILE_CHECKSUM__ = 0x6A94C23E
#__TEST2_NUM_CONNECTS__ = 1
__SMALL_FILE_NAME__ = "File1KB.htm"
__SMALL_FILE_SIZE__ = 0x400
__SMALL_FILE_CHECKSUM__ = -0x345B05EF
__TEST2_NUM_CONNECTS__ = 2000

# Test 3:
__RUN_TEST_3__ = False
#__TEST3_MAX_CONNECTS__ = 5
__TEST3_MAX_CONNECTS__ = 3000

# Test 4:
__RUN_TEST_4__ = False

# Test 5:
__RUN_TEST_5__ = False
__START_RATE__ = 0.2    # The amount of time in between starting 10-connection threads

# For test 5b, 5c, try varying the start rate to be slower and __MAX_IDLE_SECONDS__ to be lower to
# test fairness/QoS

# Test 6:
__RUN_TEST_6__ = False
__TEST6_NUM_CONNECTS__ = 5000

# Get command line arguments if they are present
if len(sys.argv) > 1:
	if int(sys.argv[1]) == 1:
		__RUN_TEST_1__ = True
	elif int(sys.argv[1]) == 2:
		__RUN_TEST_2__ = True
	elif int(sys.argv[1]) == 3:
		__RUN_TEST_3__ = True
	elif int(sys.argv[1]) == 4:
		__RUN_TEST_4__ = True
	elif int(sys.argv[1]) == 5:
		__RUN_TEST_5__ = True
	elif int(sys.argv[1]) == 6:
		__RUN_TEST_6__ = True

###############################################################
# Utility functions for test cases

# Error handling should be done by the caller of this function
def GetFile(ServerAddress,ServerPort,FileName,FileSize):
    soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    soc.connect((__SERVER_ADDRESS__,__SERVER_PORT__))
    
    # Send the HTTP request
    sendbuf = __HTTP_REQ__ % FileName
    toSend = len(sendbuf)
    sent = 0
    while sent < toSend:
        sent += soc.send(sendbuf[sent:])
        
    # Receive the file
    recvbuf = ""
    received = 0
    header = True
    while (len(recvbuf) < FileSize) or header:
        tmpbuf = soc.recv(__MAX_RECV__)
        if len(tmpbuf) == 0:
            raise socket.error
        recvbuf = recvbuf + tmpbuf
        if(header):
            pos = recvbuf.find("\r\n\r\n")
            if(pos != -1):
                recvbuf = recvbuf[pos+4:]
                header = False
    soc.close()
    return recvbuf
    
# maxBwPerConnection is in bytes / second
def DownloadThread(doneIndex, doneList, numConnections, maxBwPerConnection, fileName, fileSize):
    sockets = []
    received = { }
    headerLen = len("HTTP/1.0 200 OK\r\n\r\n")
    
    if maxBwPerConnection == 0:
        bw = 8192
    else:
        bw = int(maxBwPerConnection / 10)
    try:    
        for i in range(numConnections):
            soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            soc.connect((__SERVER_ADDRESS__,__SERVER_PORT__))
            
            # Send the HTTP request
            sendbuf = __HTTP_REQ__ % fileName
            toSend = len(sendbuf)
            sent = 0
            while sent < toSend:
                sent += soc.send(sendbuf[sent:])
            # Save the socket info
            sockets.append(soc)
            received[soc] = 0
                
        # Now receive all of the data using select
        count = 0
        idlecount = 0
        startTime = time.time()
        while True:
            (ins, _, exs) = select.select(sockets,[],sockets,1)
            if exs:
                for s in sockets:
                    s.close()
                doneList[doneIndex] = -1
                return
            if len(ins) == 0:
                idlecount = idlecount + 1
                if idlecount >= __MAX_IDLE_SECONDS__:
                    for s in sockets:
                        s.close()
                    doneList[doneIndex] = -1
                    return
            for inSock in ins:
                idlecount = 0
                data = inSock.recv(bw)
                received[inSock] = received[inSock] + len(data)
                if (len(data) == 0) or received[inSock] >= fileSize + headerLen:
                    sockets.remove(inSock)
                    inSock.close()
                    if received[inSock] < fileSize + headerLen:
                        for s in sockets:
                            s.close()
                        doneList[doneIndex] = -1
                        return
                    doneList[doneIndex] = doneList[doneIndex] + 1
                    if(len(sockets) == 0):
                        return
                # TODO: Verify checksum (This requires lots of extra client state)
            # Sleep 100 milliseconds for each iteration of the loop to regulate bandwidth
            count += 1
            if (maxBwPerConnection != 0) and (count * 0.1 + startTime > time.time()):
                time.sleep((count * 0.1 + startTime) - time.time())
    except socket.error, arg:
        for s in sockets:
            s.close()
        doneList[doneIndex] = -1
        return

###############################################################
# Test 1: Get a single 500 KB file from the web server
if __RUN_TEST_1__:
    print "###############################################################\n" + \
        "Test #1: Get a single file from the web server (Simple Correctness)"
    
    startTime = time.time()
    try:
        recvbuf = GetFile(__SERVER_ADDRESS__, __SERVER_PORT__, __SINGLE_FILE_NAME__, __SINGLE_FILE_SIZE__)
    
        # Check the file checksum
        print "Calculating checksum"
        fchecksum = zlib.crc32(recvbuf)
        #print recvbuf
        print "Received file with checksum: %08X, elapsed time: %f seconds" % \
            (fchecksum,time.time()-startTime)
        if(fchecksum == __SINGLE_FILE_CHECKSUM__):
            print "Checksum is correct (PASSED)"
        else:
            print "Checksum is incorrect! (FAILED)"
            raise
    except socket.error, arg:
        try: msg = arg[1]
        except: msg = arg
        print "Error in test #1 %s" % msg
        raise socket.errror

###############################################################
# Test 2: Repeatedly get the same small file 2000 times
if __RUN_TEST_2__:
    print "###############################################################\n" + \
        "Test #2: Get a small file %d times from the web server (Resource de-allocation and latency)" % __TEST2_NUM_CONNECTS__
    
    startTime = time.time()
    
    try:
        for i in xrange(__TEST2_NUM_CONNECTS__):
            recvbuf = GetFile(__SERVER_ADDRESS__, __SERVER_PORT__, __SMALL_FILE_NAME__, __SMALL_FILE_SIZE__)
                
            # Check the file checksum
            fchecksum = zlib.crc32(recvbuf)
            if(fchecksum != __SMALL_FILE_CHECKSUM__):
                print "Error in test #2: Incorrect checksum"
                raise
        if(fchecksum == __SMALL_FILE_CHECKSUM__ or True):
            print "Received %d files successfully, average request time: %f milliseconds" % (__TEST2_NUM_CONNECTS__,1000 * (time.time()-startTime)/__TEST2_NUM_CONNECTS__)
    except socket.error, arg:
        try: msg = arg[1]
        except: msg = arg
        print "Error in test #2 after %d connections %s" % (i,msg)
        raise socket.errror
    
    
###############################################################
# Test 3: Connect to the server until the server cannot handle more connections
if __RUN_TEST_3__:
    print "###############################################################\n" + \
        "Test #3: Connect to the server until the server cannot handle more connections (Scalability)"
    
    sockets = []
    try:
        for i in xrange(__TEST3_MAX_CONNECTS__):
            soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            soc.connect((__SERVER_ADDRESS__,__SERVER_PORT__))
            
            # Send the HTTP request
            sendbuf = __HTTP_REQ__ % __SINGLE_FILE_NAME__
            toSend = len(sendbuf)
            sent = 0
            while sent < toSend:
                sent += soc.send(sendbuf[sent:])
            sockets.append(soc)
                 
        print "Server handled all %d connections without failing" % __TEST3_MAX_CONNECTS__
        for tempsoc in sockets:
            tempsoc.close()
    except socket.error, arg:
        try: msg = arg[1]
        except: msg = arg
        print "Server accepted first %d connections successfully, died on error: %s" % (len(sockets) - 1,msg)
        for tempsoc in sockets:
            tempsoc.close()
        raise socket.error

###############################################################
# Test 4: Maximum bandwidth of 100 concurrent downloads
if __RUN_TEST_4__:
    print "###############################################################\n" + \
        "Test #4: Maximum bandwidth of 100 concurrent downloads (Throughput)"

    startTime = time.time()
    
    doneList = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    # Start 10 threads, each connecting 10 times with no(0) max bandwidth
    for i in range(10):
        thread.start_new_thread(DownloadThread,(i,doneList,10,0,__SINGLE_FILE_NAME__,__SINGLE_FILE_SIZE__))
        
    sys.stdout.write("Waiting for downloads to complete")
    count = 0
    while(doneList.count(10) + doneList.count(-1) < 10):
        time.sleep(.005)    # Sleep for 5 milliseconds
        count = count + 1
        if count == 10:
            sys.stdout.write(".")
            count = 0
    sys.stdout.write("\n")
        
        
    if doneList.count(-1) > 0:
        print "Socket error while receiving files in test case #4"
    else:    
        print "Test case #4 finished successfully -- Average bandwidth for ~50 MByte transfer: %s MBits/Second" % \
            (((100*500*8)/1024.0) / (time.time() - startTime))

###############################################################
# Test 5: Maximum number of concurrent slow downloads
if __RUN_TEST_5__:
    print "###############################################################\n" + \
        "Test #5: Maximum number of concurrent slow downloads (Fairness and scalability)"

    startTime = time.time()
    
    doneList = []
    connectionCount = 0
    sys.stdout.write("Starting concurrent slow downloads")
    while doneList.count(-1) == 0:
        doneList.append(0)
        thread.start_new_thread(DownloadThread,(connectionCount,doneList,10,1000,__SINGLE_FILE_NAME__,__SINGLE_FILE_SIZE__))
        connectionCount = connectionCount + 1
        time.sleep(__START_RATE__)    # Give each ten connections a little while to fail if they are going to
        sys.stdout.write(".")
    sys.stdout.write("\n")
        
    inProgress = len(doneList) * 10
    while doneList.count(-1) > 0:
        doneList.remove(-1)
    
    finished = sum(doneList)
        
    print "%d connections in progress when terminated, %d finished" % (inProgress-finished,finished)
    exit()
    
###############################################################
# Test 6: Repeatedly connect, send request, then close connection
if __RUN_TEST_6__:
    print "###############################################################\n" + \
        "Test #6: Repeatedly connect, send request, then close connection (Resource de-allocation and error handling)"
    
    try:
        for i in xrange(__TEST6_NUM_CONNECTS__):
            soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            soc.connect((__SERVER_ADDRESS__,__SERVER_PORT__))
            
            # Send the HTTP request
            sendbuf = __HTTP_REQ__ % __SINGLE_FILE_NAME__
            toSend = len(sendbuf)
            sent = 0
            while sent < toSend:
                sent += soc.send(sendbuf[sent:])
            
            # Close the connection
            soc.close()
        print "Successfully handled %d subsequent failed connections (PASSED)" % __TEST6_NUM_CONNECTS__
    except socket.error, arg:
        try: msg = arg[1]
        except: msg = arg
        print "Error in test #6 after %d connections %s" % (i,msg)
        raise socket.errror