#ifndef __SELECT_SERVER_H_INCLUDED__
#define __SELECT_SERVER_H_INCLUDED__

#define CONNECTION_BUF_SIZE 0x1000

#ifndef USHORT
#define USHORT unsigned short
#endif

#ifndef ULONG
#define ULONG unsigned long
#endif

#define SCANBUF_SIZE 0x400		// 1 KB, scan buffer size for sscanf()
#define BACKLOG 0x100			// Default listen backlog
#define MAX_CONNECTIONS 0x800	// Hard limit on the maximum number of connections (passed as nfds to select)

#ifndef MAX_PATH
#define MAX_PATH 256			// Change this value to whatever the max path is for your system, or define
								// it as another macro like PATH_MAX
#endif

typedef enum _connectionState
{
	StateUninitialized,
	StateWaitingForRequest,
	StateSendingFile
} connectionState;

typedef struct _connectionInfo
{
	SOCKET sock;
	connectionState state;
	char fileName[MAX_PATH];
	long filePosition;
	char buf[CONNECTION_BUF_SIZE];
	int bufpos;
	int bufsize;
} connectionInfo;


// This funciton parses an HTTP request from the client.
// Parameters:
//   request [input] - A pointer to the beginning of the HTTP request. This string
//                     should be null-terminated at the end of the request. (Search
//					   for "\r\n\r\n" sentinal at the end of the HTTP request, and
//                     set a character to NULL right after the sentinal.)
//   fileName [output] - A pointer to the buffer that will receive the file name
//   maxFileLength [input] - The size of the fileName buffer. If this value is
//							 greater than 4 KB (0x1000), then it will be set to
//                           4 KB (0x1000) instead.
// Return Value:
//   parse_request returns 0  or 1 if the header is parsed succcessfully indicating
//   version HTTP/1.0 or HTTP/1.1, respectively. If parse_request fails, then it will
//   return -1.
int parse_request(const char* request, char* fileName, unsigned int maxFileLength)
{
	char scanbuf1[0x400];
	char scanbuf2[0x1000];
	char scanbuf3[0x400];

	scanbuf1[0x3FF] = '\0';
	scanbuf2[0xFFF] = '\0';
	scanbuf3[0x3FF] = '\0';

	if(sscanf(request,"%1023s %4095s %1023s", scanbuf1, scanbuf2, scanbuf3) != 3)
		return -1;
	// Return an error if the file name is too long
	if(strlen(scanbuf2) > maxFileLength)
		return -1;
	// Copy the file name
	strncpy(fileName,scanbuf2,min(maxFileLength,0x1000));
	// Check GET and HTTP/1.x
	if(strncmp(scanbuf1,"GET",strlen("GET")))
		return -1;
	if(!strncmp(scanbuf3,"HTTP/1.0",strlen("HTTP/1.0")))
		return 0;
	else if(!strncmp(scanbuf3,"HTTP/1.1",strlen("HTTP/1.1")))
		return 1;
	else
		return -1;
}

#endif